require 'active_record'

module Netfira::WebConnect
  class Migration < ActiveRecord::Migration

    class << self
      attr_accessor :log_to_stdout
    end

    def add_localized_string(args)
      @l10n_buffer << args if @l10n_buffer
    end

    def create_table(table_name, options = {}, &block)
      super table_prefix(table_name), options, &block
    end

    def create_record_table(table_name, options = {})
      Refinements.running_migration = self
      @l10n_buffer = options[:with_l10n] && []
      options[:writable] ||= options[:sendable]
      create_table table_name, options do |t|
        t.string options[:origin_key] || :"#{table_name.to_s.singularize}_id", index: true unless options[:writable]
        yield t
        t.references :shop, index: true
        create_tree_table t if options[:tree]
        create_writable_table t if options[:writable]
        create_sendable_table t if options[:sendable]
        t.binary :digest, limit: 16
        t.timestamps
        t.datetime :deleted_at
        t.index :deleted_at
      end
      if @l10n_buffer
        create_l10n_table table_name do |t|
          @l10n_buffer.each do |args|
            t.string *args
          end
        end
      end
      add_table_def table_name, options
    end

    def create_l10n_table(table_name)
      create_table l10n_suffix(table_name) do |t|
        t.references table_name.to_s.singularize.to_sym, index: true
        t.string :language, limit: 20, index: true
        yield t
      end
    end

    def create_file_table(table_name)
      create_record_table table_name, origin_key: :file_name, file: true do |t|
        t.string :remote_location
        t.string :locale
        yield t if block_given?
        t.integer :size
        t.binary :checksum, limit: 16
      end
    end

    def create_relation_table(first, second)
      first, second = [first, second].sort
      create_table :"#{first}_to_#{second}" do |t|
        t.references first.to_s.singularize, second.to_s.singularize, index: true
        t.timestamps
        t.datetime :deleted_at
        t.index :deleted_at
      end
    end

    def add_table_def(table_name, options)
      Models::Table.create options.select{|k| %i[origin_key writable sendable file].include? k}.merge name: table_name
    end

    def write(text = '')
      if verbose
        puts text if Migration.log_to_stdout
        Netfira::WebConnect.logger.info text
      end
    end

    def schema_migrations_table_name
      Netfira::WebConnect.schema_migrations_table_name
    end

    private

    def l10n_suffix(name = '')
      Netfira::WebConnect.db_table_l10n_suffix name
    end

    def table_prefix(name = '')
      Netfira::WebConnect.db_table_prefix name
    end

    def create_tree_table(t)
      t.string :parent_id
      t.index [:shop_id, :parent_id]
    end

    def create_writable_table(t)
      t.binary :guid, limit: 16
      t.index :guid, unique: true
    end

    def create_sendable_table(t)
      t.integer :delivery_status, default: 0, limit: 3, index: true
    end

  end
end

require_relative 'migration/refinements'
require_relative 'migration/base_methods'
