module Netfira::WebConnect

  def self.migrate
    force_db_connection { ActiveRecord::Migrator.migrate migrations_dir }
  end

  def self.rollback
    force_db_connection { ActiveRecord::Migrator.rollback migrations_dir }
  end

  def self.needs_migration?
    original = ActiveRecord::Migrator.migrations_paths
    ActiveRecord::Migrator.migrations_paths = [migrations_dir]
    result = force_db_connection { ActiveRecord::Migrator.needs_migration? }
    ActiveRecord::Migrator.migrations_paths = original
    result
  end

  private

  def self.migrations_dir
    @migrations_dir ||= ROOT.join('web_connect/db_schema').to_s
  end

  # This method tricks ActiveRecord into using the connection for Netfira::WebConnect::Model
  # instead of the ActiveRecord::Base connection, which it's hard-wired to do for migrations
  # and other tasks. Two methods are backed up and replaced, then restored after yield. Those
  # methods are identical to the originals, except they specify `Model` in place of `self`.
  def self.force_db_connection

    # Save the original methods
    originals = [:connection_pool, :retrieve_connection, :schema_migrations_table_name].map do |name|
      [name, ActiveRecord::Base.method(name)]
    end.to_h

    # Save the original schema migration table
    migration_table = ActiveRecord::SchemaMigration.all.table
    original_migration_table = [migration_table.name, migration_table.engine]

    # Sub in our own migration table details
    migration_table.name = Netfira::WebConnect.schema_migrations_table_name
    migration_table.engine = Model

    # Add some new methods
    class << ActiveRecord::Base
      def connection_pool
        connection_handler.retrieve_connection_pool(Model) or raise ActiveRecord::ConnectionNotEstablished
      end
      def retrieve_connection
        connection_handler.retrieve_connection(Model)
      end
      def schema_migrations_table_name
        Netfira::WebConnect.schema_migrations_table_name
      end
    end

    # Run the given block
    result = yield

    # Restore the original migration table name
    migration_table.name, migration_table.engine = original_migration_table

    # Restore the original methods
    originals.each { |name, method| ActiveRecord::Base.define_singleton_method name, method }

    # Return the result of the given block
    result
  end

end
