require_relative 'record/relations'
require_relative 'record/materialization'
require_relative 'record/serializer'
require_relative 'record/digests'
require_relative 'record/sendable'
require_relative 'record/writable'
require_relative 'record/events'

module Netfira::WebConnect
  class Model::Record
    extend Materialization
    include Relations
    include Serializer
    include Digests
    include Events

    class << self

      def sendable
        return super unless self == Model::Record # This should raise an error!
        @sendable_classes ||= Models.constants.map do |name|
          Models.const_get name
        end.select do |klass|
          Class === klass && klass < Model::Record && klass.sendable?
        end
      end

      def table_name
        @table_name ||= if self == Model::Record
          Models::Table.table_name
        else
          Netfira::WebConnect.db_table_prefix(plural_name).to_s
        end
      end

      def has_languages?
        const_defined? :Translation, false
      end

      def tree?
        @is_tree ||= attribute_names.include? 'parent_id'
      end

      def origin_key
        @origin_key ||= writable? ? :guid : (table_props.origin_key || "#{single_name}_id").to_sym
      end

      def writable?
        @writable ||= table_props.writable
      end

      def sendable?
        @sendable ||= table_props.sendable
      end

      def find_by_origin_id(shop, id)
        id = id.b if OctaWord === id
        with_deleted.find_by origin_key => id, shop_id: shop.id
      end

      def find_or_initialize_by_origin_id(shop, id)
        id = Guid.from_base64(id).b if origin_key == :guid
        find_by_origin_id(shop, id).tap { |record| record.restore! if record && record.destroyed? } or
            new(origin_key => id, shop_id: shop.id)
      end

      def find_or_create_by_origin_id(shop, id)
        record = find_or_initialize_by_origin_id(shop, id)
        record.save unless record.persisted?
        record
      end

      def has_file?
        @has_file ||= self < FileRecord
      end

      def schema
        @schema ||= Netfira::WebConnect.schema[self]
      end

      def related_classes
        @related_classes ||= []
      end
    end

    after_initialize :init

    %i[has_languages? tree? origin_key writable? sendable? has_file?].each do |method|
      define_method(method) { self.class.__send__ method }
    end

    def origin_id
      __send__ origin_key
    end

    def as_readonly
      readonly!
      yield
      instance_variable_set :@readonly, false
      self
    end

    def as_readwrite
      instance_variable_set :@readonly, false
      yield
      readonly!
      self
    end

    def reload
      clear_translations if has_languages?
      super
    end

    def write_with_type_casting(key, value)
      schema = self.class.schema
      if schema.localize.include? key
        set_translated_string key, value
      elsif key == :parent_id
        value = value.to_s
        write_attribute key, value.empty? ? nil : value
      else
        type = schema.columns[key]
        return false unless type
        value = value.to_s if type == 'string'
        write_attribute key, value
      end
      true
    end

    private

    def self.table_props
      @table_props
    end

    def init
      self.shop_id ||= 0
    end

    def after_add_relation(model)
      model.shop_id = shop_id
    end

  end
end

require_relative 'record/file_record'
require_relative 'record/translated_string'
require_relative 'record/translation'
require_relative 'record/translations'
require_relative 'record/tree'
