module Netfira::WebConnect
  class Model::Record
    module Events
      extend ActiveSupport::Concern

      included do
        before_save :dispatch_before_update
        around_save :dispatch_around_update
        after_save :dispatch_after_update
        before_destroy :dispatch_before_destroy
        around_destroy :dispatch_around_destroy
        after_destroy :dispatch_after_destroy
      end

      private

      def update_event_args
        @event_args ||= [].tap do |args|
          previous = persisted? && changed_attributes.map do |key, value|
            # TODO: solve this type casting issue without guessing by attribute names!
            value = Checksum.new(value) if value and (key == 'checksum' or key == 'digest')
            [key, value]
          end.to_h
          event_name = previous ? 'update' : 'create'
          args << ["#{event_name}_#{self.class.single_name}", self]
          args << [event_name, self]
          if previous
            previous.merge! previous_translations if has_languages?
            args.each { |v| v << previous }
          end
        end
      end

      def destroy_event_args
        @event_args ||= [
            ["delete_#{self.class.single_name}", self],
            ['delete', self]
        ]
      end

      def dispatch_with_args(args, *timing)
        as_readonly do
          if block_given?
            dispatch_event timing, *args.first do
              dispatch_event(timing, *args.last) { as_readwrite { yield } }
            end
          else
            args.each { |x| dispatch_event timing, *x }
          end
        end
        @event_args = nil if timing.include? :after
        true
      end

      def dispatch_before_update
        dispatch_with_args update_event_args, :before
      end

      def dispatch_around_update
        dispatch_with_args(update_event_args, :around) { yield }
      end

      def dispatch_after_update
        # Paperclip sets this up for us, but it runs after this handler, which
        # means delegates don't get a saved image. Boo. We can save it twice,
        # because the second save (the one Paperclip set up) will no-op.
        attachment.save if has_file?
        dispatch_with_args update_event_args, :on, :after
      end

      def dispatch_before_destroy
        dispatch_with_args destroy_event_args, :before, :on
      end

      def dispatch_around_destroy
        dispatch_with_args(destroy_event_args, :around) { yield }
      end

      def dispatch_after_destroy
        dispatch_with_args destroy_event_args, :after
      end

    end
  end
end
