require 'paperclip'
require 'open-uri'

Paperclip.interpolates :shop_id do |record, x|
  record.instance.shop_id
end

Paperclip.interpolates :model_type do |record, x|
  record.instance.class.name.demodulize.underscore.pluralize
end

module Netfira::WebConnect
  class Model::Record::FileRecord
    include Paperclip::Glue

    class << self

      def table_name
        self == Model::Record::FileRecord ? Models::Table.table_name : super
      end

      def origin_key
        :file_name
      end

    end

    class Validator < ActiveModel::Validator
      def self.remote_file_sources_filter
        @remote_file_sources_filter ||= Regexp.new(ENV['REMOTE_FILE_SOURCES_FILTER'] || '^https?://', Regexp::IGNORECASE)
      end
      def validate(record)
        if record.remote_location && !(record.remote_location =~ self.class.remote_file_sources_filter)
          record.errors[:remote_location] << 'Unacceptable format'
        end
      end
    end

    validates_with Validator

    after_initialize :set_file_defaults

    def <<(input)
      if input.is_a? Model::Record
        super input
      else
        self.data = input
      end
    end

    def checksum
      bin = self[:checksum]
      @checksum ||= bin && Checksum.new(bin)
    end

    def checksum=(value)
      @checksum = nil
      self[:checksum] = normalize_checksum(value)
    end

    def data
      @stream ||= if remote_location
        open remote_location.sub(/^file:/, '')
      elsif attachment.present?
        Paperclip.io_adapters.for attachment
      end
    end

    def data=(value)
      @stream = nil
      self.attachment = value
      value.rewind
      self.checksum = Checksum.of(value.read)
    end

    def attachment_file_name
      file_name
    end

    def attachment_file_name=(value)
      self.file_name = value unless value.nil?
    end

    def attachment_file_size
      size
    end

    def attachment_file_size=(value)
      self.size = value unless value.nil?
    end

    def url
      exists? && remote_location || attachment.url
    end

    def exists?
      !!(remote_location || attachment.exists?)
    end
    alias_method :exist?, :exists?

    private

    def set_file_defaults
      self.locale ||= shop.locale
    end

    def normalize_checksum(value)
      if Checksum === value
        value.b
      elsif String === value
        case value.length
          when 22 then OctaWord.from_base64(value).b
          when 32 then OctaWord.from_hex(value).b
          else value
        end
      else
        value
      end
    end

  end
end
