module Netfira::WebConnect
  class Model::Record

    # This module is mixed into records that have translations.
    module Translations

      # Called when getting a translated string, e.g. product.description
      def get_translated_string(key)
        translated_string_for key
      end

      # Called when assigning a translated string, e.g. product.description = 'Stuff'
      def set_translated_string(key, value)
        if Hash === value
          replace_translations_for key, value
        else
          translated_string_for(key)[nil] = value
        end
      end

      def locale
        shop.locale
      end

      def [](key)
        if translated_attribute_names.include? key.to_s
          get_translated_string key.to_sym
        else
          super key
        end
      end

      def []=(key, value)
        if translated_attribute_names.include? key.to_s
          set_translated_string key.to_sym, value
        else
          super key, value
        end
      end

      private

      def translations
        @translations ||= translation_models.map{ |model| [model.language, model]}.to_h
      end

      def previous_translations
        Hash.new { |hash, k| hash[k] = {} }.tap do |result|
          translations.each do |language, translation|
            translation.changed_attributes.each do |attr, value|
              result[attr][language] = value if translation.class.translated_attribute_names.include? attr
            end
          end
        end
      end

      def translated_strings
        @translated_strings ||= {}
      end

      def translation_for_language(lang)
        lang = lang.to_s
        translations[lang] ||= translation_models.build(language: lang)
      end

      def replace_translations_for(key, value)
        translations.each { |lang, model| model[key] = value[lang] }
        translated_string_for(key).merge! value
      end

      def translated_string_for(key)
        translated_strings[key] ||= begin
          t = Model::Record::TranslatedString.new
          t.on(:get) { |lang|        translation_for_language(lang || locale).__send__ key }
          t.on(:set) { |lang, value| translation_for_language(lang || locale).__send__ :"#{key}=", value }
          t.on(:all) { translations.map{ |lang, model| [lang, model[key]] }.reject{ |x| x[1].nil? }.to_h }
          t
        end
      end

      def clear_translations
        @translations = nil
        @translated_strings = nil
        self
      end

      def translated_attribute_names
        self.class::Translation.translated_attribute_names
      end

    end
  end
end
