module Netfira::WebConnect
  # noinspection RubyStringKeysInHashInspection
  class Model::Record

    module Writable
      extend ActiveSupport::Concern

      included do

        # Add a model initializer that generates a random GUID
        after_initialize :create_guid

        # Attributes that can be sent
        def self.attributes_to_send
          @attributes_to_send ||= attribute_names +
              (has_languages? ? self::Translation.translated_attribute_names : []) -
              Serializer::EXCLUDE_FROM_SERIALIZE
        end

      end

      def to_deliverable
        attributes_as_camel_cased_hash.tap do |result|

          # Include GUIDs for sendables, e.g. "orderId" => "hd8jdi38jds02893hdo8wq3"
          result["#{self.class.single_name_camel}Id"] = guid.to_s if sendable?

          # Included timestamps, e.g. "createdAt" => "2014-09-27T11:10:00+10:00"
          %w(created_at updated_at).each do |field|
            result[field.camelize :lower] = __send__(field).strftime '%FT%T%:z'
          end

          # Include collections, e.g. "lines" => [{...}, ...]
          result.merge! self.class.schema.has_many.keys.map { |k| [k, __send__(k).map(&:to_deliverable)] }.to_h
        end
      end

      # Used by older downstream methods, e.g. newOrders, newInboundDocuments etc
      def to_deprecated_downstream
        fields = attributes_as_camel_cased_hash(true)
        if sendable?
          sendable_fields = {
              "#{self.class.single_name_camel}GUID" => guid.to_s,
              "#{self.class.single_name_camel}Id" => id.to_s,
              finalised: '1'
          }
          fields.merge! sendable_fields
        end
        {created: created_at.to_i, fields: fields}.merge deprecated_downstream_children
      end

      def guid
        @guid ||= Guid.from(self[:guid])
      end

      def guid=(_)
        raise 'GUIDs are read-only'
      end

      private

      def attributes_as_camel_cased_hash(stringify = false)
        self.class.attributes_to_send.map do |key|
          value = __send__(key)
          value = value.to_f if BigDecimal === value
          value = value.to_s if stringify
          [key.camelize(:lower), value]
        end.to_h
      end

      def deprecated_downstream_children
        self.class.schema.has_many.keys.map do |key|
          [key, __send__(key).map(&:to_deprecated_downstream)]
        end.to_h
      end

      def create_guid
        self[:guid] ||= (@guid = Guid.create).b
      end

      module ClassMethods
        def single_name_camel
          @single_name_camel ||= single_name.camelize :lower
        end
      end

    end

  end
end
