module Netfira::WebConnect
  class RackApp::Action
    include RackApp::Exceptions::HttpExceptions

    VALID_BASE64 = %r`^([a-z\d+/]{4})*([a-z\d+/]{4}|[a-z\d+/]{3}=|[a-z\d+/]{2}==)$`i

    def import_env(env)
      @env = env
      @shop = nil
      session = nil

      # Parse the environment
      request = Rack::Request.new env

      # Restore sessions
      session_lifetime = Netfira::WebConnect.session_lifetime
      session_token = session_lifetime && env['HTTP_X_SESSION']
      if session_token
        session = Models::Session.by_token(session_token)
        if session
          @shop = session.shop
        else
          raise Unauthorized, 'Session has expired'
        end
      end

      # Authentication
      unless @shop
        authenticator = Netfira::WebConnect.authenticator
        if authenticator.respond_to? :call
          shop_name = env['HTTP_X_SHOP_NAME']
          password = request['pw'] || env['HTTP_X_PASSWORD']

          # Basic auth
          auth = Rack::Auth::Basic::Request.new(env)
          if auth.provided? && auth.basic?
            shop_name ||= auth.username
            password ||= auth.credentials[1]
          end

          # Make shop_name a copy, so the authenticator can't mutate
          # the original string
          shop_name = shop_name.dup if shop_name

          result = authenticator.call shop_name, password
          raise Unauthorized unless result
          @shop = Netfira::WebConnect::Models::Shop.find_or_create_by(name: shop_name)

          header :x_vary_password, result if String === result

          # Sessions
          if session_lifetime
            session ||= @shop.sessions.new
            session.expires_at = Time.now + session_lifetime if Fixnum === session_lifetime
            session.save
            header :x_session, session.token
          end

        elsif authenticator.nil?
          @shop = Netfira::WebConnect.anonymous_shop
        else
          raise 'Authenticator is not callable'
        end
      end

      # The request verb (PUT, GET, POST etc)
      @verb = request.request_method.downcase.to_sym

      # Query string
      @query_string = request.GET

      # The X-Timeout header
      timeout = env['HTTP_X_TIMEOUT']
      @timeout = timeout.to_i if timeout

      # Path components
      if env['PATH_INFO'] =~ /\A\/\d+\/[^\/]+\/(.+)\z/
        @path = $1.split('/').map{ |x| Rack::Utils.unescape x }
      end

      # Input
      if put? or post?
        @input = request.body

        # Decode base64
        if (env['HTTP_CONTENT_ENCODING'] || env['CONTENT_ENCODING'] || '').downcase == 'base64'
          input = @input.read.gsub(/\s+/, '')
          raise BadRequest, 'Invalid base64 in request body' unless input.length % 4 == 0 && input =~ VALID_BASE64
          @input = StringIO.new(input.unpack('m').first)
        end

        # Unserialize JSON
        begin
          @input = JSON.parse @input.read.sub(/\A\z/, 'null'), quirks_mode: true if request.media_type == 'application/json'
        rescue JSON::ParserError => error
          raise BadRequest.new('Invalid JSON in request body', details: error.message.sub(/^\d+:\s*/, ''))
        end
      end

    end

  end
end
